import sys
from io import StringIO
from pathlib import Path

import click
import paramiko
from click.testing import CliRunner
from paramiko import SSHClient
from paramiko.sftp_client import SFTPClient


def generate_path(client: SSHClient, target_path: str) -> Path:
    if target_path.startswith("/"):
        return Path(target_path)
    # If target path doesn't start with "/" it means it's a relative path
    _stdin, stdout, _stderr = client.exec_command("pwd")
    return Path(stdout.readline().strip()) / Path(target_path)


def create_target_dir(client: SSHClient, target_folder_path: Path):
    _stdin, _stdout, _stderr = client.exec_command(f"mkdir -p {target_folder_path}")


def copy_file_to_server_helper(
    client: SSHClient,
    sftp: SFTPClient,
    source_path: Path,
    target_path: Path,
    create_target_folder: bool = True,
):
    if create_target_folder:
        create_target_dir(client, target_path.parent)
    return sftp.put(str(source_path.absolute()), str(target_path.absolute()))


@click.command()
@click.option("--host", default="", help="host address")
@click.option("--port", default=22, help="port")
@click.option("--username", default="", help="user name")
@click.option("--password", default="", help="user password")
@click.option("--pkey", default="", help="private key")
@click.option("--sourcepath", default="", help="source file to copy")
@click.option("--targetpath", default="", help="target path to copy the file to")
@click.option("--createtargetdir", default=True, help="create directory if file doesnt exist")
def copy_file_to_server(
    host: str,
    port: int,
    username: str,
    password: str,
    pkey: str,
    sourcepath: str,
    targetpath: str,
    createtargetdir: bool = True,
):
    with SSHClient() as client:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        pkey_loaded = paramiko.RSAKey.from_private_key(StringIO(pkey))
        client.connect(
            hostname=host,
            port=port,
            username=username,
            password=password,
            pkey=pkey_loaded,
        )

        path_source = Path(sourcepath)
        path_target = generate_path(client, target_path=targetpath)

        with client.open_sftp() as sftp:
            assert path_source.is_file()
            print(f"Copying {path_source.absolute()} to {path_target.absolute()}")
            copy_file_to_server_helper(
                client,
                sftp,
                path_source,
                path_target,
                create_target_folder=createtargetdir,
            )
    sys.exit(0)


def main():
    runner = CliRunner()
    ip = "some.url"
    username = "some_name"
    key = "my ssh key"
    result = runner.invoke(
        copy_file_to_server,
        [
            f"--host={ip}",
            f"--username={username}",
            f"--pkey={key}",
            "--sourcepath=copy_file_to_server.py",
            "--targetpath=test2/copy_file_to_server3.py",
        ],
    )
    for line in result.output.splitlines():
        print(line)
    if result.exit_code != 0:
        print(result.exception)
        print(result.exit_code)


if __name__ == "__main__":
    main()



import subprocess
import sys
from io import StringIO
from pathlib import Path
from typing import Optional, Set

import click
import paramiko
from click.testing import CliRunner
from paramiko import SSHClient
from paramiko.sftp_client import SFTPClient

from burny_common.copy_file_to_server import copy_file_to_server_helper, generate_path


def copy_folder_to_server_helper(
    client: SSHClient,
    sftp: SFTPClient,
    path_source_root_folder: Path,
    path_target_root_folder: Path,
    file_or_folder: Path,
    allowed_files: Optional[Set] = None,
):
    source_path_relative_to_root = file_or_folder.relative_to(path_source_root_folder)
    if file_or_folder.is_file():
        if allowed_files is not None and file_or_folder.absolute() not in allowed_files:
            # print(f"Ignoring file/folder {file_or_folder}")
            return
        # print(f"Copying file {file_or_folder}")
        path_target = path_target_root_folder / source_path_relative_to_root
        copy_file_to_server_helper(client, sftp, file_or_folder, path_target, create_target_folder=True)
    else:
        for file_or_folder2 in file_or_folder.iterdir():
            # Recursively copy folders
            copy_folder_to_server_helper(
                client,
                sftp,
                path_source_root_folder,
                path_target_root_folder,
                file_or_folder2,
                allowed_files=allowed_files,
            )


@click.command()
@click.option("--host", default="", help="host address")
@click.option("--port", default=22, help="port")
@click.option("--username", default="", help="user name")
@click.option("--password", default="", help="user password")
@click.option("--pkey", default="", help="private key")
@click.option("--sourcepath", default="", help="source folder to copy")
@click.option("--targetpath", default="", help="which folder the output should be")
@click.option("--respectgitignore", default=True, help="ignore files that are ignored by .gitignore")
def copy_folder_to_server(
    host: str,
    port: int,
    username: str,
    password: str,
    pkey: str,
    sourcepath: str,
    targetpath: str,
    respectgitignore: bool = True,
):
    client: SSHClient
    with SSHClient() as client:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        pkey_loaded = paramiko.RSAKey.from_private_key(StringIO(pkey))
        client.connect(hostname=host, port=port, username=username, password=password, pkey=pkey_loaded)

        path_source_root_folder = Path(sourcepath)
        path_target_root_folder = generate_path(client, targetpath)

        allowed_files: Optional[Set] = None
        if respectgitignore:
            proc = subprocess.Popen(
                ["git", "ls-files"],
                cwd=path_source_root_folder.absolute(),
                stdout=subprocess.PIPE,
            )
            output = proc.stdout
            if output is not None:
                files = output.read().decode()
                allowed_files = {(path_source_root_folder / Path(file)).absolute() for file in files.splitlines()}

        with client.open_sftp() as sftp:
            assert path_source_root_folder.is_dir()
            copy_folder_to_server_helper(
                client,
                sftp,
                path_source_root_folder,
                path_target_root_folder,
                # Start copying this folder
                path_source_root_folder,
                allowed_files=allowed_files,
            )
    sys.exit(0)


def main():
    runner = CliRunner()
    ip = "some.url"
    username = "some_name"
    key = "my ssh key"
    result = runner.invoke(
        copy_folder_to_server,
        [
            f"--host={ip}",
            f"--username={username}",
            f"--pkey={key}",
            "--sourcepath=/home/burny/github/python-template",
            "--targetpath=test5",
        ],
    )
    for line in result.output.splitlines():
        print(line)
    if result.exit_code != 0:
        print(result.exception)
        print(result.exit_code)


if __name__ == "__main__":
    main()



import time
from contextlib import contextmanager

from loguru import logger


@contextmanager
def time_this(label: str):
    start = time.perf_counter_ns()
    yield
    end = time.perf_counter_ns()
    logger.info(f"TIME {label}: {(end - start) / 1e9} sec")


# Use like this
if __name__ == "__main__":
    with time_this("square rooting"):
        for i in range(10**4):
            _x = i**0.5



from __future__ import annotations

import math


class Minheap:
    def __init__(self):
        """Implementation of binary heap as min-heap"""
        self.heap: list[int] = [-1]

    def __repr__(self):
        return_list = []
        multiples_of_two = {2**n for n in range(1, 1 + int(math.log(len(self.heap), 2)))}
        for i, value in enumerate(self.heap[1:], start=1):
            if i in multiples_of_two:
                return_list.append("\n")
            return_list.append(value)
            return_list.append(" ")
        return "".join(str(x) for x in return_list)

    def get_parent(self, index: int) -> int:
        return self.heap[index // 2]

    @classmethod
    def get_parent_index(cls, index: int) -> int:
        return index // 2

    @classmethod
    def get_left_child_index(cls, index: int) -> int:
        return index * 2

    @classmethod
    def get_right_child_index(cls, index: int) -> int:
        return index * 2 + 1

    def get_left_child(self, index: int) -> int | None:
        try:
            return self.heap[index * 2]
        except IndexError:
            return None

    def get_right_child(self, index: int) -> int | None:
        try:
            return self.heap[index * 2 + 1]
        except IndexError:
            return None

    def is_empty(self) -> bool:
        return len(self.heap) < 2

    def _swap(self, index1: int, index2: int):
        self.heap[index1], self.heap[index2] = self.heap[index2], self.heap[index1]

    def _move_up(self, number: int, index: int):
        if index <= 1:
            return
        parent = self.get_parent(index)
        if parent > number:
            parent_index = self.get_parent_index(index)
            self._swap(parent_index, index)
            self._move_up(number, parent_index)

    def _move_down(self, number: int, index: int):
        # Swap position with the smallest child
        left = self.get_left_child(index)
        right = self.get_right_child(index)
        # Has no children
        if left is None:
            return
        # Has only left child, so try to swap with that if it is smaller than 'number'
        if right is None:
            if number > left:
                self._swap(self.get_left_child_index(index), index)
            return
        # Has both children, check which child is the smallest, swap if smallest child is smaller than 'number'
        left_index = self.get_left_child_index(index)
        right_index = self.get_right_child_index(index)
        smallest_child_index = left_index if left < right else right_index
        smallest_child = self.heap[smallest_child_index]
        if smallest_child < number:
            self._swap(smallest_child_index, index)
            self._move_down(number, smallest_child_index)

    def insert(self, number: int):
        self.heap.append(number)
        self._move_up(len(self.heap) - 1, len(self.heap) - 1)

    def get_min(self) -> int:
        if len(self.heap) > 0:
            return self.heap[1]
        raise IndexError("get_min from empty heap")

    def delete_min(self):
        # Swap minimum with last item in list before removing
        if len(self.heap) < 2:
            raise IndexError("delete_min from empty heap")
        if len(self.heap) < 3:
            self.heap.pop(1)
            return
        last_item_index = len(self.heap) - 1
        self._swap(1, last_item_index)
        # Move the item at index 1 down to its proper position
        self.heap.pop()
        self._move_down(self.heap[1], 1)

    def build(self, my_list: list):
        self.heap = [-1]
        for i in my_list:
            self.insert(i)


if __name__ == "__main__":
    p = Minheap()
    build_list = [1, 2, 3, 4, 5, 6, 7]
    p.build(build_list)
    assert len(p.heap) == 8, "build() function or insert() function not working as expected"
    """ p:
    1
    2 3
    4 5 6 7
    """
    for i in build_list:
        assert not p.is_empty(), "Min heap should be not empty, but is returned to be empty"
        value = p.get_min()
        assert value == i, (
            f"get_min or delete_min function not working as expected, received value '{value}' "
            f"but should have been '{i}', heap:\n{p}"
        )
        p.delete_min()

    assert p.is_empty(), "Min heap should be empty, but isn't"



import sys
from io import StringIO

import click
import paramiko
from click.testing import CliRunner
from paramiko import SSHClient


@click.command()
@click.option("--host", default="", help="host address")
@click.option("--port", default=22, help="port")
@click.option("--username", default="", help="user name")
@click.option("--password", default="", help="user password")
@click.option("--pkey", default="", help="private key")
@click.option("--command", default="ls -la", help="command to execute")
def run_command_on_server(host: str, port: int, username: str, password: str, pkey: str, command: str):
    client: SSHClient
    with SSHClient() as client:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        pkey_loaded = paramiko.RSAKey.from_private_key(StringIO(pkey))
        client.connect(hostname=host, port=port, username=username, password=password, pkey=pkey_loaded)
        _stdin, stdout, _stderr = client.exec_command(f"{command}")
        lines = stdout.readlines()
        for line in lines:
            print(line)
    sys.exit(0)


def main():
    runner = CliRunner()
    ip = "some.url"
    username = "some_name"
    key = "my ssh key"
    result = runner.invoke(
        run_command_on_server,
        [
            f"--host={ip}",
            f"--username={username}",
            f"--pkey={key}",
            "--command=ls -lah",
        ],
    )
    for line in result.output.splitlines():
        print(line)
    if result.exit_code != 0:
        print(result.exception)
        print(result.exit_code)


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

from loguru import logger


class NodeOutcome(Enum):
    NULL = 0
    SUCCESS = 1
    FAIL = 2
    RUNNING = 3


@dataclass
class ActionNode:
    child_actions: list[ActionNode] = field(default_factory=list)
    continue_on_success: bool = True
    continue_on_fail: bool = False

    def run(self) -> NodeOutcome:
        outcome = self.run_node()
        if outcome == NodeOutcome.RUNNING:
            return NodeOutcome.RUNNING
        if outcome == NodeOutcome.FAIL:
            return NodeOutcome.FAIL

        for child_node in self.child_actions:
            child_outcome = child_node.run()
            if child_outcome == NodeOutcome.RUNNING:
                return NodeOutcome.RUNNING

            if child_outcome == NodeOutcome.SUCCESS and not self.continue_on_success:
                return outcome

            if child_outcome == NodeOutcome.FAIL and not self.continue_on_fail:
                return outcome

        return outcome

    def run_node(self) -> NodeOutcome:
        """Run your code here"""
        _self = self
        return NodeOutcome.SUCCESS

    def print_tree(self):
        """TODO: output tree to console"""

    def display_tree(self):
        """TODO: output tree to display/monitor via matplotlib or graphviz or similar"""


if __name__ == "__main__":

    class MyAction(ActionNode):
        my_status: int = 0

        def run_node(self) -> NodeOutcome:
            self.my_status += 1
            logger.info(f"Counting up: {self.my_status}")
            if self.my_status != 10:
                return NodeOutcome.RUNNING
            return NodeOutcome.SUCCESS

    some_action = MyAction()
    my_root_node = ActionNode([some_action])

    while 1:
        result = my_root_node.run()
        if result == NodeOutcome.SUCCESS:
            break



from pathlib import Path
from typing import Iterable


def convert_string_to_path(multi_line_string: str) -> Iterable[Path]:
    """
    Convert a given list of paths as single string

        /root/some/file
        /root/some/other/file

    to an iterator of Path
    """
    for line in multi_line_string.strip().split("\n"):
        yield Path(line)


def recurse_path(path: Path, depth: int = 0) -> Iterable[Path]:
    """
    Go through a given path recursively and return file paths

    If depth == 0: only allow file path
    If depth == 1: if given a folder, return containing file paths
    Depth > 1 allow recursively to go through folders up to a given depth
    """
    if path.is_file():
        yield path
    elif path.is_dir() and depth > 0:
        for subfile_path in sorted(path.iterdir()):
            yield from recurse_path(subfile_path, depth=depth - 1)
